package ec.gpxml.engine;

import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Reader;
import java.util.ArrayList;

import org.jsefa.Deserializer;
import org.jsefa.xml.XmlIOFactory;
import org.jsefa.xml.XmlReaderFactory;
import org.jsefa.xml.XmlSerializer;

/**
 * This class is the core for the <b>XMLGP</b> tool.<br/>
 * This class has the methods to serialize, deserialize and print XML.
 * 
 * @author dharani 
 *
 */
public final class XMLGP
{
	/**
	 * Loads XML from the input stream, converts that into project object and returns the object.
	 * @param dataInputStream input data stream which has XML content
	 * @return project - the object which is the holder for the content of XML
	 */
	public static Project deserializeFromStream(DataInputStream dataInputStream) 
	{
		Project project = null;
		
		try
		{	
			//gets the project object from the input stream
			project = deserializeProject(dataInputStream);
		}
    	catch(Exception e)
    	{
    		e.printStackTrace();
    	}
    	    	
    	return project;
	}	
	
	/**
	 * Loads XML from an input file, converts XML into project object and returns the object.
	 * @param inputXmlFileName the name of the input XML file
	 * @return project - the object which is the holder for the content of XML
	 */
	public static Project deserializeFromFile(String inputXmlFileName) 
	{
		Project project = null;
		
		FileInputStream fis = null;
    	BufferedInputStream bis = null;
    	DataInputStream dataInputStream = null;
    			
		try
		{
			//converts file input stream into data input stream
			fis = new FileInputStream(inputXmlFileName);
	        bis = new BufferedInputStream(fis);
	        dataInputStream = new DataInputStream(bis);
	        
	        //gets the project object from the stream
	        project = deserializeProject(dataInputStream);
		}
		catch(FileNotFoundException fnfe)
    	{
    		fnfe.printStackTrace();
    	}
    	catch(Exception e)
    	{
    		e.printStackTrace();
    	}
    	finally
    	{
    		try
    		{  
    			//close all the opened input streams
    			dataInputStream.close();
    			bis.close();
    			fis.close();   			 	 
    		}
    		catch(IOException ioe)
    		{
    			ioe.printStackTrace();
    		}
    	}  
    	
    	return project;
	}
	
	
	/**
	 * Deserializes the input stream into the project object.
	 * @param dataInputStream input data stream which has XML content
	 * @return project - the object which is the holder for the content of XML
	 */
    private static Project deserializeProject(DataInputStream dataInputStream) 
    {  
    	Project project = null;
    	
    	Deserializer deserializer = null;    	
    	Reader fileReader = null;
    	    	
    	try
    	{   
    		//calls jsefa to deserialize into project
	        fileReader = XmlReaderFactory.create(dataInputStream);    		
	        deserializer = XmlIOFactory.createFactory(Project.class).createDeserializer();
	        deserializer.open(fileReader);  
	                    
	        while (deserializer.hasNext()) 
	        {           	
	        	project = deserializer.next();	        	
	        }         
    	}    	
    	catch(Exception e)
    	{
    		e.printStackTrace();
    	}
    	finally
    	{
    		try
    		{      			
    			deserializer.close(true);    			 			 	 
    		}
    		catch(Exception e)
    		{
    			e.printStackTrace();
    		}
    	}   
    	
    	return project;
    }    
         
    
    /**
     * Serializes the XML content to the stream.
     * @param project the object which is the holder for the content of XML
     */
    public static void serializeToStream(Project project)
	{		
		XmlSerializer serializer = null;
		PrintWriter printWriter = null;
		        
        try
		{    
        	//calls jsefa to serialize project to stream
        	serializer = XmlIOFactory.createFactory(Project.class).createSerializer();
        	printWriter = new PrintWriter(System.out, true);
            serializer.open(printWriter);
            serializer.write(project);            
            serializer.flush();            
            serializer.close(true);  
            serializer = null;
            System.out.toString();
            //printWriter.flush();
            //printWriter.close();        	
		}
        catch(Exception e)
        {
        	e.printStackTrace();
        }
        finally 
		{
			try			
			{
				if(null != serializer)
				{
					serializer.flush();
					serializer.close(true);	
				}
				if(null != printWriter)
				{
					printWriter.flush();
					printWriter.close();				
				}		
				
			} 
			catch (Exception e) 
			{
				e.printStackTrace();
			}
		}		
	}
	
    /**
     * Serializes the XML content to the file.
     * @param project the object which is the holder for the content of XML     
     * @param outputXmlFileName the name of the output XML file
     */  
    public static void serializeToFile(Project project, String outputXmlFileName)
	{
		XmlSerializer serializer = null;
		FileWriter fstream = null;
		BufferedWriter writer = null;
		
		try
		{
			//calls jsefa to serialize project to stream
			serializer = XmlIOFactory.createFactory(Project.class).createSerializer();
			fstream = new FileWriter(outputXmlFileName);
			writer = new BufferedWriter(fstream);			
			serializer.open(writer);
			serializer.write(project);
			serializer.flush();
			serializer.close(true);
			serializer = null;
		}
		catch (Exception e) 
		{
			e.printStackTrace();
		} 
		finally 
		{
			try			
			{
				if(null != serializer)
				{
					serializer.flush();
					serializer.close(true);
				}
				if(null != writer)
				{
					writer.close();				
				}
				if(null != fstream)
				{
					fstream.close();
				}
			} 
			catch (Exception e) 
			{
				e.printStackTrace();
			}
		}		
	}
           
    
    /**
     * Prints project object which holds the XML content. 
     * This is for debugging purpose.
     * @param project the object which is the holder for the content of XML
     */
    public static void printGPResults(Project project) 
	{
		try 
		{
			ArrayList<Statistics> statList;
			
			System.out.println("Project ID: " + project.getId());
			System.out.println("Project Name: " + project.getName());
			System.out.println("Project description: " + project.getDescription());
			System.out.println("Project Date: " + project.getDate());
			System.out.println("Project Org: " + project.getOrganization());
			System.out.println("Project Fund: " + project.getFund());
			
			//print the statistics of project
			statList = (ArrayList<Statistics>) project.getStatistics();
			printStatistics(statList);

			for (Experiment experiment : project.getExperiments()) 
			{
				System.out.println("Experiment ID: " + experiment.getId());
				System.out.println("Experiment Description: " + experiment.getDescription());
				System.out.println("Experiment Date: " + experiment.getDate());
				statList = (ArrayList<Statistics>) experiment.getStatistics();
				printStatistics(statList);	
				
				for (Treatment treatment : experiment.getTreatments()) 
				{
					System.out.println("Treatment ID:" + treatment.getId());
					statList = (ArrayList<Statistics>) treatment.getStatistics();
					printStatistics(statList);
					
					for (Trial trial : treatment.getTrials()) 
					{
						System.out.println("Trial ID: " + trial.getId());
						statList = (ArrayList<Statistics>) trial.getStatistics();
						printStatistics(statList);
						
						for (Generation generation : trial.getGenerations()) 
						{
							System.out.println("Gen ID: " + generation.getId());
							statList = (ArrayList<Statistics>) generation.getStatistics();
							printStatistics(statList);
							
							for (Individual individual : generation.getIndividuals()) 
							{
								System.out.println("Individual ID: " + individual.getId());
								statList = (ArrayList<Statistics>) individual.getStatistics();
								printStatistics(statList);
								
								for (Representation representation : individual.getRepresentations()) 
								{
									System.out.println("Representation ID: " + representation.getId());
									System.out.println("Representation Type: " + representation.getType());
									statList = (ArrayList<Statistics>) representation.getStatistics();
									printStatistics(statList);
									
									Node node = representation.getNode();
									//System.out.println("Node ID: " + node.getId());
									System.out.println("Node Type: " + node.getType());
									System.out.println("Node Value: " + node.getValue());
									
									//print the children of this node
									printChildren(node);
								}
							}
						}
					}
				}
			}
		} 
		catch (Exception e) 
		{
			e.printStackTrace();
		}
	}
    
    /**
     * Prints the children of the given node.
     * @param node
     */
    private static void printChildren(Node node) 
	{
    	ArrayList<Node> children = (ArrayList<Node>) node.getChildren();
    	
    	if(null != children)
		{    		
			for (Node child : children) 
			{
				if(null != child)
				{	
					//System.out.println("child Node ID: "	+ child.getId());
					System.out.println("child Node Type: " + child.getType());
					System.out.println("child Node Value: " + child.getValue());					
					
					printChildren(child);
				}
			}	
		}    	
	}
    
    /**
     * Prints statistics of Project, Experiment, Treatment, Trial, Generation,
     * Individual and Representation
     * @param statList list of statistics
     */
    private static void printStatistics(ArrayList<Statistics> statList) 
	{
    	if(null != statList)
		{
			for (Statistics stat : statList) 
			{
				if(null != stat)
				{
					System.out.println("Project stat: " + stat.getName()					
							+ ":" + stat.getValue());
				}
			}
		}
	}
}